@@ -188,11 +188,7 @@ impl<'a> Iterator for Lifetimes<'a> {
     type Item = &'a LifetimeParam;
 
     fn next(&mut self) -> Option<Self::Item> {
-        let next = match self.0.next() {
-            Some(item) => item,
-            None => return None,
-        };
-        if let GenericParam::Lifetime(lifetime) = next {
+        if let GenericParam::Lifetime(lifetime) = self.0.next()? {
             Some(lifetime)
         } else {
             self.next()
@@ -206,11 +202,7 @@ impl<'a> Iterator for LifetimesMut<'a> {
     type Item = &'a mut LifetimeParam;
 
     fn next(&mut self) -> Option<Self::Item> {
-        let next = match self.0.next() {
-            Some(item) => item,
-            None => return None,
-        };
-        if let GenericParam::Lifetime(lifetime) = next {
+        if let GenericParam::Lifetime(lifetime) = self.0.next()? {
             Some(lifetime)
         } else {
             self.next()
@@ -224,11 +216,7 @@ impl<'a> Iterator for TypeParams<'a> {
     type Item = &'a TypeParam;
 
     fn next(&mut self) -> Option<Self::Item> {
-        let next = match self.0.next() {
-            Some(item) => item,
-            None => return None,
-        };
-        if let GenericParam::Type(type_param) = next {
+        if let GenericParam::Type(type_param) = self.0.next()? {
             Some(type_param)
         } else {
             self.next()
@@ -242,11 +230,7 @@ impl<'a> Iterator for TypeParamsMut<'a> {
     type Item = &'a mut TypeParam;
 
     fn next(&mut self) -> Option<Self::Item> {
-        let next = match self.0.next() {
-            Some(item) => item,
-            None => return None,
-        };
-        if let GenericParam::Type(type_param) = next {
+        if let GenericParam::Type(type_param) = self.0.next()? {
             Some(type_param)
         } else {
             self.next()
@@ -260,11 +244,7 @@ impl<'a> Iterator for ConstParams<'a> {
     type Item = &'a ConstParam;
 
     fn next(&mut self) -> Option<Self::Item> {
-        let next = match self.0.next() {
-            Some(item) => item,
-            None => return None,
-        };
-        if let GenericParam::Const(const_param) = next {
+        if let GenericParam::Const(const_param) = self.0.next()? {
             Some(const_param)
         } else {
             self.next()
@@ -278,11 +258,7 @@ impl<'a> Iterator for ConstParamsMut<'a> {
     type Item = &'a mut ConstParam;
 
     fn next(&mut self) -> Option<Self::Item> {
-        let next = match self.0.next() {
-            Some(item) => item,
-            None => return None,
-        };
-        if let GenericParam::Const(const_param) = next {
+        if let GenericParam::Const(const_param) = self.0.next()? {
             Some(const_param)
         } else {
             self.next()
@@ -368,7 +344,7 @@ generics_wrapper_impls!(Turbofish);
 #[cfg(feature = "printing")]
 impl<'a> TypeGenerics<'a> {
     /// Turn a type's generics like `<X, Y>` into a turbofish like `::<X, Y>`.
-    pub fn as_turbofish(&self) -> Turbofish {
+    pub fn as_turbofish(&self) -> Turbofish<'a> {
         Turbofish(self.0)
     }
 }
@@ -538,7 +514,9 @@ ast_struct! {
 #[cfg(feature = "parsing")]
 pub(crate) mod parsing {
     use crate::attr::Attribute;
-    use crate::error::{self, Result};
+    #[cfg(feature = "full")]
+    use crate::error;
+    use crate::error::{Error, Result};
     use crate::ext::IdentExt as _;
     use crate::generics::{
         BoundLifetimes, ConstParam, GenericParam, Generics, LifetimeParam, PredicateLifetime,
@@ -693,14 +671,7 @@ pub(crate) mod parsing {
                 lifetimes: {
                     let mut lifetimes = Punctuated::new();
                     while !input.peek(Token![>]) {
-                        let attrs = input.call(Attribute::parse_outer)?;
-                        let lifetime: Lifetime = input.parse()?;
-                        lifetimes.push_value(GenericParam::Lifetime(LifetimeParam {
-                            attrs,
-                            lifetime,
-                            colon_token: None,
-                            bounds: Punctuated::new(),
-                        }));
+                        lifetimes.push_value(input.parse()?);
                         if input.peek(Token![>]) {
                             break;
                         }
@@ -739,12 +710,8 @@ pub(crate) mod parsing {
                     }
                     bounds.push_value({
                         let allow_precise_capture = false;
-                        let allow_tilde_const = true;
-                        TypeParamBound::parse_single(
-                            input,
-                            allow_precise_capture,
-                            allow_tilde_const,
-                        )?
+                        let allow_const = true;
+                        TypeParamBound::parse_single(input, allow_precise_capture, allow_const)?
                     });
                     if !input.peek(Token![+]) {
                         break;
@@ -776,8 +743,8 @@ pub(crate) mod parsing {
     impl Parse for TypeParamBound {
         fn parse(input: ParseStream) -> Result<Self> {
             let allow_precise_capture = true;
-            let allow_tilde_const = true;
-            Self::parse_single(input, allow_precise_capture, allow_tilde_const)
+            let allow_const = true;
+            Self::parse_single(input, allow_precise_capture, allow_const)
         }
     }
 
@@ -785,14 +752,12 @@ pub(crate) mod parsing {
         pub(crate) fn parse_single(
             input: ParseStream,
             #[cfg_attr(not(feature = "full"), allow(unused_variables))] allow_precise_capture: bool,
-            allow_tilde_const: bool,
+            allow_const: bool,
         ) -> Result<Self> {
             if input.peek(Lifetime) {
                 return input.parse().map(TypeParamBound::Lifetime);
             }
 
-            let begin = input.fork();
-
             #[cfg(feature = "full")]
             {
                 if input.peek(Token![use]) {
@@ -810,6 +775,8 @@ pub(crate) mod parsing {
                 }
             }
 
+            let begin = input.fork();
+
             let content;
             let (paren_token, content) = if input.peek(token::Paren) {
                 (Some(parenthesized!(content in input)), &content)
@@ -817,24 +784,11 @@ pub(crate) mod parsing {
                 (None, input)
             };
 
-            let is_tilde_const =
-                cfg!(feature = "full") && content.peek(Token![~]) && content.peek2(Token![const]);
-            if is_tilde_const {
-                let tilde_token: Token![~] = content.parse()?;
-                let const_token: Token![const] = content.parse()?;
-                if !allow_tilde_const {
-                    let msg = "`~const` is not allowed here";
-                    return Err(error::new2(tilde_token.span, const_token.span, msg));
-                }
-            }
-
-            let mut bound: TraitBound = content.parse()?;
-            bound.paren_token = paren_token;
-
-            if is_tilde_const {
-                Ok(TypeParamBound::Verbatim(verbatim::between(&begin, input)))
-            } else {
+            if let Some(mut bound) = TraitBound::do_parse(content, allow_const)? {
+                bound.paren_token = paren_token;
                 Ok(TypeParamBound::Trait(bound))
+            } else {
+                Ok(TypeParamBound::Verbatim(verbatim::between(&begin, input)))
             }
         }
 
@@ -842,11 +796,11 @@ pub(crate) mod parsing {
             input: ParseStream,
             allow_plus: bool,
             allow_precise_capture: bool,
-            allow_tilde_const: bool,
+            allow_const: bool,
         ) -> Result<Punctuated<Self, Token![+]>> {
             let mut bounds = Punctuated::new();
             loop {
-                let bound = Self::parse_single(input, allow_precise_capture, allow_tilde_const)?;
+                let bound = Self::parse_single(input, allow_precise_capture, allow_const)?;
                 bounds.push_value(bound);
                 if !(allow_plus && input.peek(Token![+])) {
                     break;
@@ -857,7 +811,7 @@ pub(crate) mod parsing {
                     || input.peek(Token![?])
                     || input.peek(Lifetime)
                     || input.peek(token::Paren)
-                    || input.peek(Token![~]))
+                    || (allow_const && (input.peek(token::Bracket) || input.peek(Token![const]))))
                 {
                     break;
                 }
@@ -869,8 +823,37 @@ pub(crate) mod parsing {
     #[cfg_attr(docsrs, doc(cfg(feature = "parsing")))]
     impl Parse for TraitBound {
         fn parse(input: ParseStream) -> Result<Self> {
+            let allow_const = false;
+            Self::do_parse(input, allow_const).map(Option::unwrap)
+        }
+    }
+
+    impl TraitBound {
+        fn do_parse(input: ParseStream, allow_const: bool) -> Result<Option<Self>> {
+            let mut lifetimes: Option<BoundLifetimes> = input.parse()?;
+
+            let is_conditionally_const = cfg!(feature = "full") && input.peek(token::Bracket);
+            let is_unconditionally_const = cfg!(feature = "full") && input.peek(Token![const]);
+            if is_conditionally_const {
+                let conditionally_const;
+                let bracket_token = bracketed!(conditionally_const in input);
+                conditionally_const.parse::<Token![const]>()?;
+                if !allow_const {
+                    let msg = "`[const]` is not allowed here";
+                    return Err(Error::new(bracket_token.span.join(), msg));
+                }
+            } else if is_unconditionally_const {
+                let const_token: Token![const] = input.parse()?;
+                if !allow_const {
+                    let msg = "`const` is not allowed here";
+                    return Err(Error::new(const_token.span, msg));
+                }
+            }
+
             let modifier: TraitBoundModifier = input.parse()?;
-            let lifetimes: Option<BoundLifetimes> = input.parse()?;
+            if lifetimes.is_none() && matches!(modifier, TraitBoundModifier::Maybe(_)) {
+                lifetimes = input.parse()?;
+            }
 
             let mut path: Path = input.parse()?;
             if path.segments.last().unwrap().arguments.is_empty()
@@ -882,12 +865,26 @@ pub(crate) mod parsing {
                 path.segments.last_mut().unwrap().arguments = parenthesized;
             }
 
-            Ok(TraitBound {
-                paren_token: None,
-                modifier,
-                lifetimes,
-                path,
-            })
+            if lifetimes.is_some() {
+                match modifier {
+                    TraitBoundModifier::None => {}
+                    TraitBoundModifier::Maybe(maybe) => {
+                        let msg = "`for<...>` binder not allowed with `?` trait polarity modifier";
+                        return Err(Error::new(maybe.span, msg));
+                    }
+                }
+            }
+
+            if is_conditionally_const || is_unconditionally_const {
+                Ok(None)
+            } else {
+                Ok(Some(TraitBound {
+                    paren_token: None,
+                    modifier,
+                    lifetimes,
+                    path,
+                }))
+            }
         }
     }
 
@@ -929,8 +926,15 @@ pub(crate) mod parsing {
     #[cfg_attr(docsrs, doc(cfg(feature = "parsing")))]
     impl Parse for WhereClause {
         fn parse(input: ParseStream) -> Result<Self> {
+            let where_token: Token![where] = input.parse()?;
+
+            if choose_generics_over_qpath(input) {
+                return Err(input
+                    .error("generic parameters on `where` clauses are reserved for future use"));
+            }
+
             Ok(WhereClause {
-                where_token: input.parse()?,
+                where_token,
                 predicates: {
                     let mut predicates = Punctuated::new();
                     loop {
@@ -1017,11 +1021,11 @@ pub(crate) mod parsing {
                             }
                             bounds.push_value({
                                 let allow_precise_capture = false;
-                                let allow_tilde_const = true;
+                                let allow_const = true;
                                 TypeParamBound::parse_single(
                                     input,
                                     allow_precise_capture,
-                                    allow_tilde_const,
+                                    allow_const,
                                 )?
                             });
                             if !input.peek(Token![+]) {
@@ -1046,13 +1050,16 @@ pub(crate) mod parsing {
             let mut params = Punctuated::new();
             loop {
                 let lookahead = input.lookahead1();
-                params.push_value(if lookahead.peek(Lifetime) || lookahead.peek(Ident) {
-                    input.parse::<CapturedParam>()?
-                } else if lookahead.peek(Token![>]) {
-                    break;
-                } else {
-                    return Err(lookahead.error());
-                });
+                params.push_value(
+                    if lookahead.peek(Lifetime) || lookahead.peek(Ident) || input.peek(Token![Self])
+                    {
+                        input.parse::<CapturedParam>()?
+                    } else if lookahead.peek(Token![>]) {
+                        break;
+                    } else {
+                        return Err(lookahead.error());
+                    },
+                );
                 let lookahead = input.lookahead1();
                 params.push_punct(if lookahead.peek(Token![,]) {
                     input.parse::<Token![,]>()?
@@ -1079,13 +1086,54 @@ pub(crate) mod parsing {
             let lookahead = input.lookahead1();
             if lookahead.peek(Lifetime) {
                 input.parse().map(CapturedParam::Lifetime)
-            } else if lookahead.peek(Ident) {
-                input.parse().map(CapturedParam::Ident)
+            } else if lookahead.peek(Ident) || input.peek(Token![Self]) {
+                input.call(Ident::parse_any).map(CapturedParam::Ident)
             } else {
                 Err(lookahead.error())
             }
         }
     }
+
+    pub(crate) fn choose_generics_over_qpath(input: ParseStream) -> bool {
+        // Rust syntax has an ambiguity between generic parameters and qualified
+        // paths. In `impl <T> :: Thing<T, U> {}` this may either be a generic
+        // inherent impl `impl<T> ::Thing<T, U>` or a non-generic inherent impl
+        // for an associated type `impl <T>::Thing<T, U>`.
+        //
+        // After `<` the following continuations can only begin generics, not a
+        // qualified path:
+        //
+        //     `<` `>`                  - empty generic parameters
+        //     `<` `#`                  - generic parameters with attribute
+        //     `<` LIFETIME `>`         - single lifetime parameter
+        //     `<` (LIFETIME|IDENT) `,` - first generic parameter in a list
+        //     `<` (LIFETIME|IDENT) `:` - generic parameter with bounds
+        //     `<` (LIFETIME|IDENT) `=` - generic parameter with a default
+        //     `<` const                - generic const parameter
+        //
+        // The only truly ambiguous case is:
+        //
+        //     `<` IDENT `>` `::` IDENT ...
+        //
+        // which we disambiguate in favor of generics because this is almost
+        // always the expected one in the context of real-world code.
+        input.peek(Token![<])
+            && (input.peek2(Token![>])
+                || input.peek2(Token![#])
+                || (input.peek2(Lifetime) || input.peek2(Ident))
+                    && (input.peek3(Token![>])
+                        || input.peek3(Token![,])
+                        || input.peek3(Token![:]) && !input.peek3(Token![::])
+                        || input.peek3(Token![=]))
+                || input.peek2(Token![const]))
+    }
+
+    #[cfg(feature = "full")]
+    pub(crate) fn choose_generics_over_qpath_after_keyword(input: ParseStream) -> bool {
+        let input = input.fork();
+        input.call(Ident::parse_any).unwrap(); // `impl` or `for` or `where`
+        choose_generics_over_qpath(&input)
+    }
 }
 
 #[cfg(feature = "printing")]
@@ -1106,7 +1154,7 @@ pub(crate) mod printing {
     use crate::print::TokensOrDefault;
     use crate::token;
     use proc_macro2::TokenStream;
-    use quote::{ToTokens, TokenStreamExt};
+    use quote::{ToTokens, TokenStreamExt as _};
 
     #[cfg_attr(docsrs, doc(cfg(feature = "printing")))]
     impl ToTokens for Generics {
